/*
    grammar.cpp
    $Id: grammar.cpp,v 1.22 2015/01/27 13:30:13 daichi Exp $

*/
#include <cstdio>
#include <cmath>
#include <utility>
#include "grammar.h"
#include "gamma.h"
#include "tree.h"
#include "util.h"

using namespace std;

void
count_rules (int x, int y, int z,
	     vector<hash_map<tuple,int> > *hash)
{
	(*hash)[x][cons(y,z)]++;
}

static int
hash_count (hash_map<int,int> &hash, int w)
{
	hash_map<int,int>::iterator it;

	if (hash.empty())
		return 0;
	
	if ((it = hash.find(w)) == hash.end())
	{
		return 0;
	} else {
		return it->second;
	}
}

static int
hash_count (hash_map<tuple,int> &hash, int y, int z)
{
	hash_map<tuple,int>::iterator it;

	if (hash.empty())
		return 0;

	if ((it = hash.find(cons(y,z))) == hash.end())
	{
		return 0;
	} else {
		return it->second;
	}
}

double
Grammar::rulep (int x)
{
	return (rtotal[x] + a) / (rtotal[x] + wtotal[x] + a + b);
}

double
Grammar::expand (int x, int y, int z)
{
	int c = hash_count (rcounts[x], y, z);
	return rulep(x) * (c + gamma) / (rtotal[x] + (K - 1) * (K - 1) * gamma);
}

double
Grammar::expand (int x, int w)
{
 	int c = hash_count (wcounts[x], w);
 	return (1 - rulep(x)) * (c + eta) / (wtotal[x] + V * eta);
}

bool
Grammar::add_counts (Tree *tree)
{
	int x, y, z, w;
	
	if (tree->left && tree->right)
	{
		x = tree->category;
		y = tree->left->category;
		z = tree->right->category;

		rtotal[x]++;
		rcounts[x][cons(y,z)]++;

		add_counts (tree->left);
		add_counts (tree->right);
	}
	else {
		x = tree->category;
		w = tree->word;

		wtotal[x]++;
		wcounts[x][w]++;
	}
}

bool
Grammar::remove_counts (Tree *tree)
{
	int x, y, z, w;

	if (tree->left && tree->right)
	{
		x = tree->category;
		y = tree->left->category;
		z = tree->right->category;
		
		rtotal[x]--;
		rcounts[x][cons(y,z)]--;

		remove_counts (tree->left);
		remove_counts (tree->right);
	} else
	{
		x = tree->category;
		w = tree->word;
		
		wtotal[x]--;
		wcounts[x][w]--;
	}
}

double
Grammar::polyalik (Tree *tree)
{
	static vector<hash_map<tuple,int> > hash;
	static vector<double> rs, ws;
	int x, y, z, c;
	double lik = 0;
	double seta = eta * V;
	double sgamma = gamma * (K - 1) * (K - 1);

	if (rs.empty()) {
		rs.resize (K);
	}
	if (ws.empty()) {
		ws.resize (K);
	}
	for (x = 0; x < K; x++) {
		rs[x] = 0;
		ws[x] = 0;
	}

	if (hash.empty())
		hash.resize (K);
	else {
		for (x = 0; x < K; x++)
			hash[x].clear();
	}
	
	tree->foreach ((nodefunc)count_rules, &hash);

	for (x = 0; x < K; x++)
	{
		hash_map<tuple,int>::iterator it;

		if (!hash[x].empty())
		{
			/* counts */
			for (it = hash[x].begin(); it != hash[x].end(); it++)
			{
				y = it->first.first;
				z = it->first.second;
				c = it->second;
				
				if (z < 0) {
					ws[x] += c;
					lik += poch (eta + wcounts[x][y], c);
				} else {
					rs[x] += c;
					lik += poch (gamma + rcounts[x][cons(y,z)], c);
				}
			}
			/* first level */
			lik += poch (a + rtotal[x], rs[x])
			       + poch (b + wtotal[x], ws[x])
			       - poch (a + b + rtotal[x] + wtotal[x], rs[x] + ws[x]);
			/* second level */
			lik -= poch (sgamma + rtotal[x], rs[x]);
			lik -= poch (seta + wtotal[x], ws[x]);
		}
		
	}

	return lik;
	
}
			

double
Grammar::treelik (Tree *tree)
{
	int x, y, z, w;

	x = tree->category;
	
	if (tree->left && tree->right)
	{
		y = tree->left->category;
		z = tree->right->category;
		return log (expand (x, y, z))
			+ treelik (tree->left) + treelik (tree->right);
	} else {
		w = tree->word;
		return log (expand (x, w));
	}
}

void
Grammar::save (char *model)
{
	int x;
	FILE *fp;

	if ((fp = fopen(strconcat(model, ".grammar"),"w")) == NULL)
	{
		perror("fopen");
		return;
	}

	fprintf(fp, "%d\t%d\n", K, V);
	fprintf(fp, "%.7e\t%.7e\t%.7e\t%.7e\n", a, b, gamma, eta);
	fprintf(fp, "\n");

	/* counts */
	for (x = 1; x < K; x++)
	{
		hash_map<int,int>::iterator it;

		fprintf(fp, "[%d]\t", x);
		for (it = wcounts[x].begin(); it != wcounts[x].end(); it++)
		{
			if (it->second > 0) {
				fprintf(fp, "%d:%d ", it->first, it->second);
			}
		}
		fprintf(fp, "\n");
	}
	fprintf(fp, "\n");
	for (x = 0; x < K; x++)
	{
		hash_map<tuple,int>::iterator it;

		fprintf(fp, "[%d]\t", x);		
		for (it = rcounts[x].begin(); it != rcounts[x].end(); it++)
		{
			if (it->second > 0) {
				fprintf(fp, "%d,%d:%d ",
					it->first.first, it->first.second, it->second);
			}
		}
		fprintf(fp, "\n");
	}

	fclose (fp);

}
		
Grammar::Grammar (int K_, double gamma_, double eta_)
{
	K = K_;
	a = 1;
	b = 1;
	gamma = gamma_;
	eta   = eta_;
	wtotal.resize (K);
	rtotal.resize (K);
	wcounts.resize (K);
	rcounts.resize (K);
}

void
Grammar::test ()
{
	int x, y, z, w;
	double p, q;

	printf("K=%d, V=%d\n", K, V);
	printf("a=%lg, b=%lg, gamma=%lg, eta=%lg\n", a, b, gamma, eta);
	
	for (x = 0; x < K; x++)
	{
		p = 0; q = 0;
		for (y = 0; y < K; y++)
		{
			for (z = 0; z < K; z++)
			{
				p += expand (x, y, z);
			}
		}
		for (w = 1; w <= V; w++)
		{
			q += expand (x, w);
		}
		printf("sum of p[%d] = %lf (%lf + %lf) rulep = %lf\n", x, p + q, p, q, rulep(x));
	}
}

void
Grammar::print ()
{
	int x;
	hash_map<int,int>::iterator it;
	hash_map<tuple,int>::iterator tt;
	
	printf("K=%d, V=%d\n", K, V);
	printf("a=%lg, b=%lg, gamma=%lg, eta=%lg\n\n", a, b, gamma, eta);

	for (x = 0; x < K; x++)
	{
		printf("category %d:\n-----------\n", x);
		printf("rtotal = %d, wtotal = %d, rulep = %lg\n",
		       rtotal[x], wtotal[x], rulep(x));

		printf("rules:");
		for (tt = rcounts[x].begin(); tt != rcounts[x].end(); tt++)
		{
			printf(" (%d,%d):%d", tt->first.first, tt->first.second, tt->second);
		}
		printf("\n");
		printf("words:");
		for (it = wcounts[x].begin(); it != wcounts[x].end(); it++)
		{
			printf(" %d:%d", it->first, it->second);
		}
		printf("\n\n");
	}

}

#if 0

int
main (int argc, char *argv[])
{
	int K = 5;
	double gamma = 0.1, eta = 0.01;
	Grammar *g = new Grammar (K, gamma, eta);
	Tree *tree, *newtree;
	g->V = 5;

	tree = new Tree (0,
			 new Tree (3,
				   new Tree (1, 1),
				   new Tree (4, 2)),
			 new Tree (2,
				   new Tree (2, 3),
				   new Tree (3,
					     new Tree (1, 4),
					     new Tree (4, 5))));
	newtree = new Tree (0,
			    new Tree (4, 2),
			    new Tree (3,
				      new Tree (4, 2),
				      new Tree (2, 1)));

	g->add_counts (tree);
	
	newtree->print ();
	
	printf("treelik = %lg, polyalik = %lg\n",
	       g->treelik(newtree), g->polyalik(newtree));
	
	
}

#endif
